module redblack
  include "mpif.h"
  integer :: RED = 0, BLACK = 1
  integer :: dtype = MPI_DOUBLE_PRECISION
  integer, dimension(MPI_STATUS_SIZE) :: status
  integer :: err
  integer :: EAST = 1, SOUTH = 2, WEST = 3, NORTH = 4

contains

  subroutine solve(U, M, N, h2, niter, comm, maxerr)
    implicit none
    
    integer, intent(in) :: M, N, niter, comm
    double precision, dimension(M, N), intent(inout) :: U
    double precision, intent(in) :: h2
    double precision, dimension(M, 4) :: sbuf, rbuf
    double precision :: eps, maxerr

    integer :: i, j, iter

    double precision :: diff, tmp
    integer :: rnorth, rsouth, reast, rwest
    integer :: rank, size, dir
    integer :: dst, length

    integer, dimension(4) :: ranks, reqs

    if (M.ne.N) then
       print*, 'Must have M == N'
    endif
    
    call MPI_Barrier(comm, err)
    call MPI_Comm_rank(comm, rank, err)
    call MPI_Comm_size(comm, size, err)
    call MPI_Cart_shift(comm, 1, 1, ranks(EAST), ranks(WEST), err)
    call MPI_Cart_shift(comm, 0, 1, ranks(NORTH), ranks(SOUTH), err)

    sbuf = 0.0
    rbuf = 0.0

    call initialize_boundary_conditions(rbuf, M)
    call copy_buffers_to_array(U, M, M, rbuf)
    
    do iter=1, niter
       eps = 0.0
       call update(U, M, sbuf, rbuf, comm, ranks, reqs, h2, eps, RED)
       call update(U, M, sbuf, rbuf, comm, ranks, reqs, h2, eps, BLACK)
       call add_up_difference(comm, eps)
       if(sqrt(eps).le.maxerr) then
          print*, 'Converged', iter, maxerr, sqrt(eps)
          return
       endif
    enddo
  end subroutine solve

  subroutine update(U, M, sbuf, rbuf, comm, ranks, reqs, h2, eps, color)
    implicit none
       
    integer :: M, comm, color
    integer, dimension(4) :: ranks, reqs
    double precision :: h2
    double precision, dimension(M, 4) :: sbuf, rbuf
    double precision, dimension(M, M) :: U
    double precision :: eps

    call update_boundaries_and_copy_to_buffers(U, M, h2, eps, color, sbuf)
    call send_buffers_nonblocking(sbuf, M, ranks, comm, reqs)
    call update_interior(U, M, h2, eps, color)
    call recv_buffers_blocking(rbuf, M, ranks, comm)
    call wait_for_requests(ranks, reqs)
    call copy_buffers_to_array(U, M, M, rbuf)


    !call recv_buffers_nonblocking(rbuf, ranks, M, comm, reqs)
    !call update_local_array(U, M, M, h2, eps, color)
    !call copy_array_to_buffers(U, M, M, sbuf)
    !call send_buffers_blocking(ranks, sbuf, M, comm)
    !call wait_for_requests(ranks, reqs)
    !call copy_buffers_to_array(U, M, M, rbuf)
    
  end subroutine update

  subroutine recv_buffers_blocking(rbuf, M, ranks, comm)
    implicit none
    
    integer :: M, comm
    integer, dimension(4) :: ranks
    double precision, dimension(M, 4) :: rbuf
    integer :: dir

    do dir=1, 4
       if (ranks(dir).ne.MPI_PROC_NULL) then
          call MPI_Recv(rbuf(1:M, dir), M, dtype, ranks(dir), 0, comm, &
               status, err)
       endif
    enddo

  end subroutine recv_buffers_blocking

  subroutine send_buffers_nonblocking(sbuf, M, ranks, comm, reqs)
    implicit none
    
    integer :: M, dir, comm
    integer, dimension(4) :: ranks, reqs
    double precision, dimension(M, 4) :: sbuf, rbuf

    do dir=1, 4
       if (ranks(dir).ne.MPI_PROC_NULL) then
          call MPI_Isend(sbuf(1:M, dir), M, dtype, &
               ranks(dir), 0, comm, reqs(dir), err)
       endif
    enddo
  end subroutine send_buffers_nonblocking

  subroutine update_boundaries_and_copy_to_buffers(U, M, h2, diff, color, buf)
    implicit none
    
    integer :: M
    integer :: i, j, color
    double precision :: diff, h2, tmp
    double precision, dimension(M, M) :: U
    double precision, dimension(M, 4) :: buf
    integer :: offset

    j = 2
    offset = mod(j + color, 2)
    do i=2 + offset, M - 2 + offset, 2
       tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + U(i + 1, j) + U(i, j + 1) - h2)
       diff = diff + (U(i, j) - tmp)**2
       U(i, j) = tmp
       buf(i, EAST) = tmp
    enddo

    j = M - 1
    offset = mod(j + color, 2)
    do i=2 + offset, M - 2 + offset, 2
       tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + U(i + 1, j) + U(i, j + 1) - h2)
       diff = diff + (U(i, j) - tmp)**2
       U(i, j) = tmp
       buf(i, WEST) = tmp
    enddo

    i = 2
    offset = mod(i + color, 2)
    do j=2 + offset, M - 2 + offset, 2
       tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + U(i + 1, j) + U(i, j + 1) - h2)
       diff = diff + (U(i, j) - tmp)**2
       U(i, j) = tmp
       buf(j, NORTH) = tmp
    enddo

    i = M - 1
    offset = mod(i + color, 2)
    do j=2 + offset, M - 2 + offset, 2
       tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + U(i + 1, j) + U(i, j + 1) - h2)
       diff = diff + (U(i, j) - tmp)**2
       U(i, j) = tmp
       buf(j, SOUTH) = tmp
    enddo

  end subroutine update_boundaries_and_copy_to_buffers

  subroutine update_interior(U, M, h2, diff, color)
    implicit none
    
    integer :: M, color
    integer :: i, j, offset
    double precision :: diff, h2, tmp
    double precision, dimension(M, M) :: U
    
    do j=3, M - 2
       offset = mod(j + color, 2)
       do i=3 + offset, M - 3 + offset, 2
          tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + &
               U(i + 1, j) + U(i, j + 1) - h2)
          diff = diff + (U(i, j) - tmp)**2
          U(i, j) = tmp
       enddo
    enddo
  end subroutine update_interior

  subroutine add_up_difference(comm, eps)
    implicit none
    
    integer :: comm
    double precision :: eps
    double precision, dimension(1) :: sendbuf, recvbuf

    sendbuf(1) = eps
    call MPI_Allreduce(sendbuf, recvbuf, 1, MPI_DOUBLE_PRECISION, &
         MPI_SUM, comm, err)
    eps = recvbuf(1)
  end subroutine add_up_difference

  subroutine send_buffers_blocking(ranks, buf, M, comm)
    implicit none
    
    integer :: M, dir, dst, comm
    integer, dimension(4) :: ranks
    double precision, dimension(M, 4) :: buf

    do dir=1, 4
       dst = ranks(dir)
       if (dst.ne.MPI_PROC_NULL) then
          call MPI_Send(buf(1:M, dir), M, dtype, dst, 0, comm, err)
       endif
    enddo
    end subroutine send_buffers_blocking

  subroutine wait_for_requests(ranks, requests)
    implicit none
    
    integer :: dir, dst
    integer, dimension(4) :: ranks, requests
    do dir=1, 4
       dst = ranks(dir)
       if (dst.ne.MPI_PROC_NULL) then
          call MPI_Wait(requests(dir), status, err)
       endif
    enddo
  end subroutine wait_for_requests

  subroutine recv_buffers_nonblocking(buf, ranks, M, comm, requests)
    implicit none
    integer :: dir, M, comm
    integer, dimension(4) :: ranks, requests
    double precision, dimension(M, 4) :: buf

    do dir=1, 4
       if (ranks(dir).ne.MPI_PROC_NULL) then
          call MPI_Irecv(buf(1:M, dir), M, dtype, &
               ranks(dir), 0, comm, requests(dir), err)
       endif
    enddo
    
  end subroutine recv_buffers_nonblocking

  subroutine copy_array_to_buffers(U, M, N, buf)
    implicit none
    integer :: M, N, i, j
    double precision, dimension(M, N) :: U
    double precision, dimension(M, 4) :: buf
    
    do i=1, M
       buf(i, EAST) = U(i, 2)
       buf(i, SOUTH) = U(M - 1, i)
       buf(i, WEST) = U(i, M - 1)
       buf(i, NORTH) = U(2, i)
    enddo
  end subroutine copy_array_to_buffers

  subroutine copy_buffers_to_array(U, M, N, buf)
    implicit none
    
    integer :: M, N, i
    double precision, dimension(M, N) :: U
    double precision, dimension(M, 4) :: buf

    do i=1, M
       U(i, 1) = buf(i, EAST)
       U(M, i) = buf(i, SOUTH)
       U(i, M) = buf(i, WEST)
       U(1, i) = buf(i, NORTH)
    enddo

  end subroutine copy_buffers_to_array

  subroutine update_local_array(U, M, N, h2, diff, color)
    implicit none
    
    integer :: M, N, color
    integer :: i, j, offset
    double precision :: diff, h2, tmp
    double precision, dimension(M, N) :: U
    
    do j=2, N - 1
       offset = mod(j + color, 2)
       do i=2 + offset, M - 2 + offset, 2
          tmp = 0.25 * (U(i - 1, j) + U(i, j - 1) + &
               U(i + 1, j) + U(i, j + 1) - h2)
          diff = diff + (U(i, j) - tmp)**2
          U(i, j) = tmp
       enddo
    enddo
  end subroutine update_local_array

  subroutine read_args(filename, Mglob, Nglob, Msize, Nsize, niter, maxerr)
    implicit none
    
    character(8), intent(in) :: filename
    integer, intent(out) :: Mglob, Nglob, Msize, Nsize, niter
    double precision, intent(out) :: maxerr
    integer :: fileno = 37
    character(16) :: buf

    open(fileno, file=filename)
    read(fileno, *) Mglob, Nglob, Msize, Nsize, niter, maxerr
  end subroutine read_args

  subroutine initialize_boundary_conditions(buf, M)
    implicit none
    integer :: M, i
    double precision, dimension(M, 4) :: buf
    
    do i=1, M
       buf(i, NORTH) = 1.0
       buf(i, SOUTH) = 1.0
       buf(i, EAST) = 0.0
       buf(i, WEST) = 0.0
    enddo
  end subroutine initialize_boundary_conditions

end module redblack
